#pragma once

#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <mscoree.h>
#include <windows.h>
#include <comdef.h>
#include <cassert>

#pragma comment (lib, "mscoree.lib")

_COM_SMARTPTR_TYPEDEF(ICLRRuntimeHost, __uuidof(ICLRRuntimeHost));
_COM_SMARTPTR_TYPEDEF(ICLRControl, __uuidof(ICLRControl));
_COM_SMARTPTR_TYPEDEF(IUnknown, __uuidof(IUnknown));
_COM_SMARTPTR_TYPEDEF(IHostControl, __uuidof(IHostControl));
_COM_SMARTPTR_TYPEDEF(IHostMemoryManager, __uuidof(IHostMemoryManager));
_COM_SMARTPTR_TYPEDEF(IHostGCManager, __uuidof(IHostGCManager));

template <typename T>
class BaseImpl : T::Interface
{
private:
	volatile LONG _refCount;
public:
	virtual ULONG STDMETHODCALLTYPE AddRef()
	{
		const LONG ret = InterlockedIncrement( &_refCount );
		if( ret <= 0 )
		{
			DebugBreak();
		}
		return ret;
	}
	virtual ULONG STDMETHODCALLTYPE Release()
	{
		LONG ret = InterlockedDecrement( &_refCount );
		if( ret <= 0 )
		{
			delete this;
			ret = 0;
		}
		return (ULONG) ret;
	}
	virtual HRESULT STDMETHODCALLTYPE  QueryInterface( 
		__in   REFIID riid,
		__out void __RPC_FAR *__RPC_FAR *ppvObject)
	{
		if( riid == IID_IUnknown )
		{
			IUnknownPtr ptr;
			ptr.Attach( this, true );
			ptr.AddRef();
			*ppvObject = ptr;
			return S_OK;
		}
		else if( riid == T::GetIID() )
		{
			T ptr;
			ptr.Attach( this, true );
			ptr.AddRef();
			*ppvObject = ptr;
			return S_OK;
		}
		else
		{
			return E_NOINTERFACE;
		}
		return E_UNEXPECTED;
	}
protected:
	BaseImpl() : _refCount(0)
	{}
};

void DebugWriteLine( const wchar_t *format, ...);

class Simple;
class Report;
class ReportEx;

template<typename T>
static IHostControlPtr CreateHostControl()
{
    return NULL;
}

template<>
IHostControlPtr CreateHostControl<Simple>();

template<>
IHostControlPtr CreateHostControl<Report>();

template<>
IHostControlPtr CreateHostControl<ReportEx>();