#include "common.h"

namespace ReportImpl
{
    class HostMemoryManagerImpl : public BaseImpl<IHostMemoryManagerPtr>
    {
    public:
        HostMemoryManagerImpl()
        {
        }
        ~HostMemoryManagerImpl()
        {
            DebugWriteLine( L"HostMemoryManagerImpl Released" );
        }
        virtual HRESULT STDMETHODCALLTYPE CreateMalloc( 
            __in   DWORD dwMallocType,
            __out IHostMalloc **ppMalloc)
        {
            *ppMalloc = NULL;
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE VirtualAlloc( 
            __in    LPVOID lpAddress,
            __in    SIZE_T dwSize,
            __in    DWORD flAllocationType,
            __in    DWORD flProtect,
            __in    EMemoryCriticalLevel eCriticalLevel,
            __out  void **ppMem)
        {
            DebugWriteLine( L"VirtualAlloc / zAhXԂ̊mۂ CLR v܂" );
            const wchar_t* criticalLeveMsg = NULL;
            switch( eCriticalLevel )
            {
            case eTaskCritical:
                criticalLeveMsg = L"TaskCritical";
                break;
            case eAppDomainCritical:
                criticalLeveMsg = L"AppDomainCritical";
                break;
            case eProcessCritical:
                criticalLeveMsg = L"ProcessCritical";
                break;
            default:
                criticalLeveMsg = L"Unknown";
                break;
            }
            DebugWriteLine( L"\tCriticalLevel: %s", criticalLeveMsg );
            DebugWriteLine( L"\tReqested Address: 0x%08x", lpAddress );
            DebugWriteLine( L"\tsize: %d byte", dwSize );
            static const DWORD AllocateTypes[] = {MEM_COMMIT, MEM_PHYSICAL, MEM_RESERVE, MEM_RESET, MEM_TOP_DOWN, MEM_WRITE_WATCH};
            static const wchar_t* AllocateTypeNames[] = {L"MEM_COMMIT", L"MEM_PHYSICAL", L"MEM_RESERVE", L"MEM_RESET", L"MEM_TOP_DOWN", L"MEM_WRITE_WATCH"};
            for( int i = 0; i < sizeof(AllocateTypes)/sizeof(DWORD); ++i )
            {
                if( (flAllocationType & AllocateTypes[i]) != 0 )
                {
                    DebugWriteLine( L"\t%s", AllocateTypeNames[i] );
                }
            }

            HRESULT ret = E_FAIL;
            void* dummy;
            ppMem = (ppMem != NULL ? ppMem : &dummy);
            *ppMem = ::VirtualAlloc( lpAddress, dwSize, flAllocationType, flProtect );
            ret = (*ppMem != NULL ? S_OK : E_OUTOFMEMORY);

            DebugWriteLine( L"\t0x%08x - 0x%08x", *ppMem , (SIZE_T)*ppMem + dwSize );

            DebugWriteLine( L"\treturn %d", ret );
            return ret;
        }


        virtual HRESULT STDMETHODCALLTYPE VirtualFree( 
            __in    LPVOID lpAddress,
            __in    SIZE_T dwSize,
            __in    DWORD dwFreeType)
        {
            DebugWriteLine( L"VirtualFree / zAhXԂ̉ CLR v܂" );

            MEMORY_BASIC_INFORMATION memInfo;
            ZeroMemory( &memInfo, sizeof(memInfo) );
            ::VirtualQuery( lpAddress, &memInfo, sizeof(memInfo) );

            const BOOL ret = ::VirtualFree( lpAddress, dwSize, dwFreeType );

            DebugWriteLine( L"\t%s", ret ? L"Success" : L"Fail" );
            DebugWriteLine( L"\t0x%08x - %d byte", lpAddress, dwSize );

            static const DWORD FreeTypes[] = {MEM_DECOMMIT, MEM_RELEASE};
            static const wchar_t* FreeTypeNames[] = {L"MEM_DECOMMIT", L"MEM_RELEASE"};
            for( int i = 0; i < sizeof(FreeTypes)/sizeof(DWORD); ++i )
            {
                if( (dwFreeType & FreeTypes[i]) != 0 )
                {
                    DebugWriteLine( L"\t%s", FreeTypeNames[i] );
                }
            }

            DebugWriteLine( L"\treturn %d", ret );
            return ret ? S_OK : E_FAIL;
        }


        virtual HRESULT STDMETHODCALLTYPE VirtualQuery( 
            __in_opt LPVOID lpAddress,
            __out_bcount_part(dwLength, return)  void *lpBuffer,
            __in    SIZE_T dwLength,
            __out  SIZE_T *pResult)
        {
            DebugWriteLine( L"VirtualQuery / zAhXԂ̑擾 CLR v܂D" );

            SIZE_T dummy = 0;
            pResult = (pResult != NULL ? pResult : &dummy);
            *pResult = ::VirtualQuery( lpAddress, reinterpret_cast<PMEMORY_BASIC_INFORMATION>(lpBuffer), dwLength );
            const MEMORY_BASIC_INFORMATION* memInfo = reinterpret_cast<MEMORY_BASIC_INFORMATION*>(lpBuffer);

            DebugWriteLine( L"\t0x%08x", lpAddress );
            DebugWriteLine( L"\tAllocationBase: 0x%08x", memInfo->AllocationBase );
            static const DWORD StateTypes[] = {MEM_COMMIT, MEM_FREE, MEM_RESERVE};
            static const wchar_t* StateTypesNames[] = {L"MEM_COMMIT", L"MEM_FREE", L"MEM_RESERVE"};
            for( int i = 0; i < sizeof(StateTypes)/sizeof(DWORD); ++i )
            {
                if( (memInfo->State & StateTypes[i]) != 0 )
                {
                    DebugWriteLine( L"\t%s", StateTypesNames[i] );
                }
            }
            static const DWORD MemoryTypes[] = {MEM_IMAGE, MEM_MAPPED, MEM_PRIVATE};
            static const wchar_t* MemoryTypesNames[] = {L"MEM_IMAGE", L"MEM_MAPPED", L"MEM_PRIVATE"};
            for( int i = 0; i < sizeof(MemoryTypes)/sizeof(DWORD); ++i )
            {
                if( (memInfo->Type & MemoryTypes[i]) != 0 )
                {
                    DebugWriteLine( L"\t%s", MemoryTypesNames[i] );
                }
            }
            static const DWORD MemoryProtectionTypes[]
                = {PAGE_EXECUTE, PAGE_EXECUTE_READ, PAGE_EXECUTE_READWRITE, PAGE_EXECUTE_WRITECOPY, PAGE_NOACCESS, 
                    PAGE_READONLY, PAGE_READWRITE, PAGE_WRITECOPY, PAGE_GUARD, PAGE_NOCACHE, PAGE_WRITECOMBINE };
            static const wchar_t* MemoryProtectionTypesNames[]
                = {L"PAGE_EXECUTE", L"PAGE_EXECUTE_READ", L"PAGE_EXECUTE_READWRITE", L"PAGE_EXECUTE_WRITECOPY", L"PAGE_NOACCESS",
                    L"PAGE_READONLY", L"PAGE_READWRITE", L"PAGE_WRITECOPY", L"PAGE_GUARD", L"PAGE_NOCACHE", L"PAGE_WRITECOMBINE"};

            for( int i = 0; i < sizeof(MemoryProtectionTypes)/sizeof(DWORD); ++i )
            {
                if( (memInfo->AllocationProtect & MemoryProtectionTypes[i]) != 0 )
                {
                    DebugWriteLine( L"\t%s", MemoryProtectionTypesNames[i] );
                }
            }

            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE VirtualProtect( 
            __in    LPVOID lpAddress,
            __in    SIZE_T dwSize,
            __in    DWORD flNewProtect,
            __out  DWORD *pflOldProtect)
        {
            DebugWriteLine( L"VirtualProtect / zAhXԂ̕ی쑮ύX CLR v܂D" );
            DWORD dummy = 0;
            pflOldProtect = (pflOldProtect != NULL ? pflOldProtect : &dummy);
            BOOL ret = ::VirtualProtect( lpAddress, dwSize, flNewProtect, pflOldProtect );
            DebugWriteLine( L"\t%s", ret ? L"Success" : L"Fail" );
            DebugWriteLine( L"\t0x%08x - %d byte", lpAddress, dwSize );

            static const DWORD MemoryProtectionTypes[]
            = {PAGE_EXECUTE, PAGE_EXECUTE_READ, PAGE_EXECUTE_READWRITE, PAGE_EXECUTE_WRITECOPY, PAGE_NOACCESS, 
                PAGE_READONLY, PAGE_READWRITE, PAGE_WRITECOPY, PAGE_GUARD, PAGE_NOCACHE, PAGE_WRITECOMBINE };
            static const wchar_t* MemoryProtectionTypesNames[] = {L"PAGE_EXECUTE", L"PAGE_EXECUTE_READ", L"PAGE_NOACCESS", L"PAGE_READONLY", L"PAGE_READWRITE", L"PAGE_WRITECOPY", L"PAGE_GUARD", L"PAGE_NOCACHE", L"PAGE_WRITECOMBINE"};

            DebugWriteLine( L"\tOld" );
            for( int i = 0; i < sizeof(MemoryProtectionTypes)/sizeof(DWORD); ++i )
            {
                if( (*pflOldProtect & MemoryProtectionTypes[i]) != 0 )
                {
                    DebugWriteLine( L"\t\t%s", MemoryProtectionTypesNames[i] );
                }
            }

            DebugWriteLine( L"\tNew", ret ? L"Success" : L"Fail" );
            for( int i = 0; i < sizeof(MemoryProtectionTypes)/sizeof(DWORD); ++i )
            {
                if( (flNewProtect & MemoryProtectionTypes[i]) != 0 )
                {
                    DebugWriteLine( L"\t\t%s", MemoryProtectionTypesNames[i] );
                }
            }

            DebugWriteLine( L"\treturn %d", ret );
            return ret ? S_OK : E_FAIL;
        }

        virtual HRESULT STDMETHODCALLTYPE GetMemoryLoad( 
            __out  DWORD *pMemoryLoad,
            __out  SIZE_T *pAvailableBytes)
        {
            DebugWriteLine( L"GetMemoryLoad / CLR 烁gpʂ̖₢킹󂯂܂" );

            MEMORYSTATUS ms;
            ZeroMemory( &ms, sizeof(ms) );
            ms.dwLength = sizeof(ms);
            GlobalMemoryStatus( &ms );

            DWORD dummy1;
            SIZE_T dummy2;

            pMemoryLoad = (pMemoryLoad != NULL ? pMemoryLoad : &dummy1);
            pAvailableBytes = (pAvailableBytes != NULL ? pAvailableBytes : &dummy2);

            *pMemoryLoad = ms.dwMemoryLoad;
            *pAvailableBytes = ms.dwAvailPhys;

            DebugWriteLine( L"\tMemoryLoad: %d, AvailableBytes: %d", *pMemoryLoad, *pAvailableBytes );

            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE RegisterMemoryNotificationCallback( 
            __in    ICLRMemoryNotificationCallback *pCallback)
        {
            DebugWriteLine( L"RegisterMemoryNotificationCallback / CLR 烁ʒmIuWFNg̎擾v܂" );

            pCallback = NULL;
            return S_OK;
        }


        virtual HRESULT STDMETHODCALLTYPE NeedsVirtualAddressSpace( 
            __in    LPVOID startAddress,
            __in    SIZE_T size)
        {
            DebugWriteLine( L"NeedsVirtualAddressSpace / zAhXԂ (t@C}bsO) gp\ǂ CLR ₢킹󂯂܂" );
            DebugWriteLine( L"\t0x%08x - %d byte", startAddress, size );
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE AcquiredVirtualAddressSpace( 
            __in    LPVOID startAddress,
            __in    SIZE_T size)
        {
            DebugWriteLine( L"AcquiredVirtualAddressSpace / zAhXԂ (t@C}bsO) gpƂ CLR ʒm܂" );
            DebugWriteLine( L"\t0x%08x - %d byte", startAddress, size );
            return S_OK;
        }


        virtual HRESULT STDMETHODCALLTYPE ReleasedVirtualAddressSpace( 
            __in    LPVOID startAddress)
        {
            DebugWriteLine( L"ReleasedVirtualAddressSpace / zAhXԂ (A}bvɂ) Ƃ CLR  ʒm܂" );
            DebugWriteLine( L"\t0x%08x", startAddress );
            return S_OK;
        }

        static IHostMemoryManagerPtr Create()
        {
            return new HostMemoryManagerImpl();
        }
    };

    class HostGCManagerImpl : public BaseImpl<IHostGCManagerPtr>
    {
    private:
        HostGCManagerImpl()
        {
        }
    public:
        virtual HRESULT STDMETHODCALLTYPE ThreadIsBlockingForSuspension()
        {
            DebugWriteLine( L"GC - ThreadIsBlockingForSuspension, TID:%d", GetCurrentThreadId() );
            return S_OK;
        }
        virtual HRESULT STDMETHODCALLTYPE SuspensionStarting()
        {
            DebugWriteLine( L"GC - SuspensionStarting, TID:%d", GetCurrentThreadId() );
            return S_OK;
        }
        virtual HRESULT STDMETHODCALLTYPE SuspensionEnding(DWORD Generation)
        {
            DebugWriteLine( L"GC - SuspensionEnding, TID:%d, Generation: %d", GetCurrentThreadId(), Generation );
            return S_OK;
        }
        static IHostGCManagerPtr Create()
        {
            return new HostGCManagerImpl();
        }
    };

    class HostControlImpl : public BaseImpl<IHostControlPtr>
    {
    private:
        IUnknownPtr _appDomainManager;
        IHostMemoryManagerPtr _memManager;
        IHostGCManagerPtr _gcManager;
        HostControlImpl()
        {
            _memManager = HostMemoryManagerImpl::Create();
            _gcManager = HostGCManagerImpl::Create();
        }
    public:
        virtual HRESULT STDMETHODCALLTYPE GetHostManager( 
            __in   REFIID riid,
            __out void **ppObject)
        {
            if( riid == IID_IHostMemoryManager )
            {
                IHostMemoryManagerPtr ptr = _memManager;
                ptr.AddRef();
                *ppObject = ptr;
            }
            else if( riid == IID_IHostGCManager )
            {
                IHostGCManagerPtr ptr = _gcManager;
                ptr.AddRef();
                *ppObject = ptr;
            }
            else
            {
                *ppObject = NULL;
            }
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE SetAppDomainManager( 
            __in DWORD dwAppDomainID,
            __in ::IUnknown *pUnkAppDomainManager)
        {
            return S_OK;
        }
        static IHostControlPtr Create()
        {
            return new HostControlImpl();
        }
    };
}

template<>
IHostControlPtr CreateHostControl<Report>()
{
    using namespace ReportImpl;
    return HostControlImpl::Create();
}