#include "common.h"

namespace SimpleImpl
{

    class HostMemoryManagerImpl : public BaseImpl<IHostMemoryManagerPtr>
    {
    public:
        HostMemoryManagerImpl()
        {
        }
        ~HostMemoryManagerImpl()
        {
        }
        virtual HRESULT STDMETHODCALLTYPE CreateMalloc( 
            __in   DWORD dwMallocType,
            __out IHostMalloc **ppMalloc)
        {
            *ppMalloc = NULL;
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE VirtualAlloc( 
            __in    LPVOID lpAddress,
            __in    SIZE_T dwSize,
            __in    DWORD flAllocationType,
            __in    DWORD flProtect,
            __in    EMemoryCriticalLevel eCriticalLevel,
            __out  void **ppMem)
        {
            HRESULT result = E_FAIL;
            void* dummy;
            ppMem = (ppMem != NULL ? ppMem : &dummy);
            *ppMem = ::VirtualAlloc( lpAddress, dwSize, flAllocationType, flProtect );
            return (*ppMem != NULL ? S_OK : E_OUTOFMEMORY);
        }


        virtual HRESULT STDMETHODCALLTYPE VirtualFree( 
            __in    LPVOID lpAddress,
            __in    SIZE_T dwSize,
            __in    DWORD dwFreeType)
        {
            const BOOL ret = ::VirtualFree( lpAddress, dwSize, dwFreeType );
            return ret ? S_OK : E_FAIL;
        }


        virtual HRESULT STDMETHODCALLTYPE VirtualQuery( 
            __in_opt LPVOID lpAddress,
            __out_bcount_part(dwLength, return)  void *lpBuffer,
            __in    SIZE_T dwLength,
            __out  SIZE_T *pResult)
        {
            SIZE_T dummy = 0;
            pResult = (pResult != NULL ? pResult : &dummy);
            *pResult = ::VirtualQuery( lpAddress, reinterpret_cast<PMEMORY_BASIC_INFORMATION>(lpBuffer), dwLength );
            const MEMORY_BASIC_INFORMATION* memInfo = reinterpret_cast<MEMORY_BASIC_INFORMATION*>(lpBuffer);
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE VirtualProtect( 
            __in    LPVOID lpAddress,
            __in    SIZE_T dwSize,
            __in    DWORD flNewProtect,
            __out  DWORD *pflOldProtect)
        {
            DWORD dummy = 0;
            pflOldProtect = (pflOldProtect != NULL ? pflOldProtect : &dummy);
            BOOL ret = ::VirtualProtect( lpAddress, dwSize, flNewProtect, pflOldProtect );
            return ret ? S_OK : E_FAIL;
        }

        virtual HRESULT STDMETHODCALLTYPE GetMemoryLoad( 
            __out  DWORD *pMemoryLoad,
            __out  SIZE_T *pAvailableBytes)
        {
            MEMORYSTATUS ms;
            ZeroMemory( &ms, sizeof(ms) );
            ms.dwLength = sizeof(ms);
            GlobalMemoryStatus( &ms );

            DWORD dummy1;
            SIZE_T dummy2;

            pMemoryLoad = (pMemoryLoad != NULL ? pMemoryLoad : &dummy1);
            pAvailableBytes = (pAvailableBytes != NULL ? pAvailableBytes : &dummy2);

            *pMemoryLoad = ms.dwMemoryLoad;
            *pAvailableBytes = ms.dwAvailPhys;
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE RegisterMemoryNotificationCallback( 
            __in    ICLRMemoryNotificationCallback *pCallback)
        {
            pCallback = NULL;
            return S_OK;
        }


        virtual HRESULT STDMETHODCALLTYPE NeedsVirtualAddressSpace( 
            __in    LPVOID startAddress,
            __in    SIZE_T size)
        {
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE AcquiredVirtualAddressSpace( 
            __in    LPVOID startAddress,
            __in    SIZE_T size)
        {
            return S_OK;
        }


        virtual HRESULT STDMETHODCALLTYPE ReleasedVirtualAddressSpace( 
            __in    LPVOID startAddress)
        {
            return S_OK;
        }

        static IHostMemoryManagerPtr Create()
        {
            return new HostMemoryManagerImpl();
        }
    };

    class HostGCManagerImpl : public BaseImpl<IHostGCManagerPtr>
    {
    private:
        HostGCManagerImpl()
        {
        }
    public:
        virtual HRESULT STDMETHODCALLTYPE ThreadIsBlockingForSuspension()
        {
            return S_OK;
        }
        virtual HRESULT STDMETHODCALLTYPE SuspensionStarting()
        {
            return S_OK;
        }
        virtual HRESULT STDMETHODCALLTYPE SuspensionEnding(DWORD Generation)
        {
            return S_OK;
        }
        static IHostGCManagerPtr Create()
        {
            return new HostGCManagerImpl();
        }
    };

    class HostControlImpl : public BaseImpl<IHostControlPtr>
    {
    private:
        IUnknownPtr _appDomainManager;
        IHostMemoryManagerPtr _memManager;
        IHostGCManagerPtr _gcManager;
        HostControlImpl()
        {
            _memManager = HostMemoryManagerImpl::Create();
            _gcManager = HostGCManagerImpl::Create();
        }
    public:
        virtual HRESULT STDMETHODCALLTYPE GetHostManager( 
            __in   REFIID riid,
            __out void **ppObject)
        {
            if( riid == IID_IHostMemoryManager )
            {
                IHostMemoryManagerPtr ptr = _memManager;
                ptr.AddRef();
                *ppObject = ptr;
            }
            else if( riid == IID_IHostGCManager )
            {
                IHostGCManagerPtr ptr = _gcManager;
                ptr.AddRef();
                *ppObject = ptr;
            }
            else
            {
                *ppObject = NULL;
            }
            return S_OK;
        }

        virtual HRESULT STDMETHODCALLTYPE SetAppDomainManager( 
            __in DWORD dwAppDomainID,
            __in ::IUnknown *pUnkAppDomainManager)
        {
            return S_OK;
        }
        static IHostControlPtr Create()
        {
            return new HostControlImpl();
        }
    };
}

template<>
IHostControlPtr CreateHostControl<Simple>()
{
    using namespace SimpleImpl;

    return HostControlImpl::Create();
}